# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from chainladder.development import Development, DevelopmentBase
from chainladder import options
from chainladder.utils.utility_functions import concat
from chainladder.utils.utility_functions import num_to_nan

import numpy as np
import pandas as pd


class CaseOutstanding(DevelopmentBase):
    """A determinisic method based on outstanding case reserves.

    The CaseOutstanding method is a deterministic approach that develops
    patterns of incremental payments as a percent of previous period case
    reserves as well as patterns for case reserves as a percent of previous
    period case reserves.  Although the patterns produces by the approach
    approximate incremental payments and case outstanding, they are converted
    into comparable multiplicative patterns for usage with the various IBNR
    methods.

    .. versionadded:: 0.8.0

    Parameters
    ----------
    paid_to_incurred: tuple or list of tuples
        A tuple representing the paid and incurred ``columns`` of the triangles
        such as ``('paid', 'incurred')``
    paid_n_periods: integer, optional (default=-1)
        number of origin periods to be used in the paid pattern averages. For
        all origin periods, set paid_n_periods=-1
    case_n_periods: integer, optional (default=-1)
        number of origin periods to be used in the case pattern averages. For
        all origin periods, set paid_n_periods=-1

    Attributes
    ----------
    ldf_: Triangle
        The estimated (multiplicative) loss development patterns.
    cdf_: Triangle
        The estimated (multiplicative) cumulative development patterns.
    case_to_prior_case_: Triangle
        The case to prior case ratios used for fitting the estimator
    case_ldf_:
        The selected case to prior case ratios of the fitted estimator
    paid_to_prior_case_: Triangle
        The paid to prior case ratios used for fitting the estimator
    paid_ldf_:
        The selected paid to prior case ratios of the fitted estimator
    """

    def __init__(
        self, paid_to_incurred=None, paid_n_periods=-1, case_n_periods=-1, groupby=None
    ):
        self.paid_to_incurred = paid_to_incurred
        self.paid_n_periods = paid_n_periods
        self.case_n_periods = case_n_periods
        self.groupby = groupby

    def fit(self, X, y=None, sample_weight=None):
        """Fit the model with X.

        Parameters
        ----------
        X : Triangle
            Set of LDFs to which the munich adjustment will be applied.
        y : Ignored
        sample_weight : Ignored

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        backend = "cupy" if X.array_backend == "cupy" else "numpy"
        self.X_ = X.copy()

        self.paid_w_ = (
            Development(n_periods=self.paid_n_periods).fit(self.X_.sum(0).sum(1)).w_
        )
        self.case_w_ = (
            Development(n_periods=self.case_n_periods).fit(self.X_.sum(0).sum(1)).w_
        )

        case_to_prior_case_naned = self.case_to_prior_case_
        case_to_prior_case_naned.values = num_to_nan(self.case_to_prior_case_.values)

        self.case_ldf_ = self.case_to_prior_case_.mean(2)  # this has the wrong value
        self.case_ldf_.values = case_to_prior_case_naned.mean(axis=2).values

        paid_to_prior_case_naned = self.paid_to_prior_case_
        paid_to_prior_case_naned.values = num_to_nan(self.paid_to_prior_case_.values)

        self.paid_ldf_ = self.paid_to_prior_case_.mean(2)  # this has the wrong value
        self.paid_ldf_.values = paid_to_prior_case_naned.mean(axis=2).values

        self.ldf_ = self._set_ldf(self.X_).set_backend(backend)

        return self

    def _set_ldf(self, X):
        print("=== in _set_ldf ===")
        paid_tri = X[self.paid_to_incurred[0]]
        incurred_tri = X[self.paid_to_incurred[1]]
        case_tri = incurred_tri - paid_tri

        original_val_date = case_tri.valuation_date

        case_ldf_ = self.case_ldf_.copy()
        case_ldf_.valuation_date = pd.Timestamp(options.ULT_VAL)
        xp = case_ldf_.get_array_module()
        # Broadcast triangle shape
        case_ldf_ = case_ldf_ * case_tri.latest_diagonal / case_tri.latest_diagonal
        case_ldf_.odims = case_tri.odims
        case_ldf_.is_pattern = False
        case_ldf_.values = xp.concatenate(
            (xp.ones(list(case_ldf_.shape[:-1]) + [1]), case_ldf_.values), axis=-1
        )

        case_ldf_.ddims = case_tri.ddims
        case_ldf_.valuation_date = case_ldf_.valuation.max()
        case_ldf_ = case_ldf_.dev_to_val().set_backend(self.case_ldf_.array_backend)

        # Will this work for sparse?
        forward = case_ldf_[case_ldf_.valuation > original_val_date].values
        forward[xp.isnan(forward)] = 1.0
        forward = xp.cumprod(forward, -1)
        1 / case_ldf_[case_ldf_.valuation <= original_val_date]

        backward = 1 / case_ldf_[case_ldf_.valuation <= original_val_date].values
        backward[xp.isnan(backward)] = 1.0
        backward = xp.cumprod(backward[..., ::-1], -1)[..., ::-1][..., 1:]
        nans = case_ldf_ / case_ldf_
        case_ldf_.values = xp.concatenate(
            (backward, (case_tri.latest_diagonal * 0 + 1).values, forward), -1
        )
        case_tri = (
            (case_ldf_ * nans.values * case_tri.latest_diagonal.values)
            .val_to_dev()
            .iloc[..., : len(case_tri.ddims)]
        )
        ld = (
            case_tri[case_tri.valuation == X.valuation_date]
            .sum("development")
            .sum("origin")
        )
        ld = ld / ld
        patterns = (1 - np.nan_to_num(X.nan_triangle[..., 1:])) * (
            self.paid_ldf_ * ld
        ).values
        paid = case_tri.iloc[..., :-1] * patterns
        paid.ddims = case_tri.ddims[1:]
        paid.valuation_date = pd.Timestamp(options.ULT_VAL)
        # Create a full triangle of incurrds to support a multiplicative LDF
        paid = (paid_tri.cum_to_incr() + paid).incr_to_cum()
        inc = (
            case_tri[case_tri.valuation > X.valuation_date]
            + paid[paid.valuation > X.valuation_date]
            + incurred_tri
        )
        # Combined paid and incurred into a single object
        paid.columns = [self.paid_to_incurred[0]]
        inc.columns = [self.paid_to_incurred[1]]
        cols = X.columns[
            X.columns.isin([self.paid_to_incurred[0], self.paid_to_incurred[1]])
        ]

        dev = concat((paid, inc), 1)[list(cols)]
        # Convert the paid/incurred to multiplicative LDF
        dev = (dev.iloc[..., -1] / dev).iloc[..., :-1]
        dev.valuation_date = pd.Timestamp(options.ULT_VAL)
        dev.ddims = X.link_ratio.ddims
        dev.is_pattern = True
        dev.is_cumulative = True

        self.case = case_tri
        self.paid = paid
        print("=== self.case ===\n", self.case)
        print("=== self.paid ===\n", self.paid)
        print("=== set LDF return ===\n", dev.cum_to_incr())
        return dev.cum_to_incr()

    @property
    def case_to_prior_case_(self):
        paid_tri = self.X_[self.paid_to_incurred[0]]
        incurred_tri = self.X_[self.paid_to_incurred[1]]

        if self.groupby is not None:
            paid_tri = paid_tri.groupby(self.groupby).sum()
            incurred_tri = incurred_tri.groupby(self.groupby).sum()

        out = (
            (incurred_tri - paid_tri).iloc[..., 1:]
            * self.case_w_
            / (incurred_tri - paid_tri).iloc[..., :-1].values
        )
        out.is_pattern = True
        out.is_cumulative = False

        return out

    @property
    def paid_to_prior_case_(self):
        paid_tri = self.X_[self.paid_to_incurred[0]]
        incurred_tri = self.X_[self.paid_to_incurred[1]]

        if self.groupby is not None:
            paid_tri = paid_tri.groupby(self.groupby).sum()
            incurred_tri = incurred_tri.groupby(self.groupby).sum()

        out = (
            paid_tri.cum_to_incr().iloc[..., 1:]
            * self.paid_w_
            / (incurred_tri - paid_tri).iloc[..., :-1].values
        )
        out.is_pattern = True

        return out

    def transform(self, X):
        """If X and self are of different shapes, align self to X, else
        return self.

        Parameters
        ----------
        X : Triangle
            The triangle to be transformed

        Returns
        -------
            X_new : New triangle with transformed attributes.
        """
        X_new = X.copy()
        X_new.ldf_ = self._set_ldf(X_new).set_backend(self.ldf_.array_backend)
        X_new._set_slicers()
        X_new.paid_ldf_ = self.paid_ldf_
        X_new.case_ldf_ = self.case_ldf_
        return X_new



import chainladder as cl

def test_trend1(clrd):
    tri = clrd[['CumPaidLoss', 'EarnedPremDIR']].sum()
    assert (
        cl.CapeCod(.05).fit(tri['CumPaidLoss'], sample_weight=tri['EarnedPremDIR'].latest_diagonal).ibnr_ ==
        cl.CapeCod().fit(cl.Trend(.05).fit_transform(tri['CumPaidLoss']), sample_weight=tri['EarnedPremDIR'].latest_diagonal).ibnr_)

def test_trend2(raa):
    tri = raa
    assert abs(
        cl.Trend(trends=[.05, .05], dates=[(None, '1985'), ('1985', None)], axis='origin').fit(tri).trend_*tri -
        tri.trend(.05, axis='origin')).sum().sum() < 1e-6



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from chainladder.methods import Benktander
import warnings


class CapeCod(Benktander):
    """Applies the CapeCod technique to triangle **X**

    Parameters
    ----------
    trend: float (default=0.0)
        The cape cod trend assumption.  Any Trend transformer on X will
        override this argument.
    decay: float (defaut=1.0)
        The cape cod decay assumption
    n_iters: int, optional (default=1)
        Number of iterations to use in the Benktander model.
    apriori_sigma: float, optional (default=0.0)
        Standard deviation of the apriori.  When used in conjunction with the
        bootstrap model, the model samples aprioris from a lognormal
        distribution using this argument as a standard deviation.
    random_state: int, RandomState instance or None, optional (default=None)
        Seed for sampling from the apriori distribution.  This is ignored when
        using as a deterministic method.
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by np.random.
    groupby:
        An option to group levels of the triangle index together for the 
        purposes of deriving the apriori measures.  If omitted, each level of 
        the triangle index will receive its own apriori computation.


    Attributes
    ----------
    triangle:
        returns **X**
    ultimate_:
        The ultimate losses per the method
    ibnr_:
        The IBNR per the method
    apriori_:
        The trended apriori vector developed by the Cape Cod Method
    detrended_apriori_:
        The detrended apriori vector developed by the Cape Cod Method
    """

    def __init__(self, trend=0, decay=1, n_iters=1, apriori_sigma=0.0,
                 random_state=None, groupby=None):
        self.trend = trend
        self.decay = decay
        self.n_iters = n_iters
        self.apriori_sigma = apriori_sigma
        self.random_state = random_state
        self.groupby = groupby

    def fit(self, X, y=None, sample_weight=None):
        """Fit the model with X.

        Parameters
        ----------
        X: Triangle-like
            Loss data to which the model will be applied.
        y: None
            Ignored
        sample_weight: Triangle-like
            The exposure to be used in the method.
        Returns
        -------
        self: object
            Returns the instance itself.
        """

        if sample_weight is None:
            raise ValueError("sample_weight is required.")
        self.apriori = 1.0
        self.X_ = self.validate_X(X)
        self.validate_weight(X, sample_weight)
        sample_weight = sample_weight.set_backend(self.X_.array_backend)
        self.apriori_, self.detrended_apriori_ = \
            self._get_capecod_aprioris(self.X_, sample_weight)
        self.expectation_ =  sample_weight * self.detrended_apriori_
        super().fit(X, y, self.expectation_)
        return self

    def _get_capecod_aprioris(self, X, sample_weight):
        """ Private method to establish CapeCod Apriori """

        if X.is_cumulative == False:
            X = X.sum('development').val_to_dev()
        latest = X.latest_diagonal
        len_orig = sample_weight.shape[-2]
        reported_exposure = sample_weight / self._align_cdf(X.copy(), sample_weight)
        reported_exposure = reported_exposure.set_backend(latest.array_backend)
        if self.groupby is not None:
            latest = latest.groupby(self.groupby).sum()
            reported_exposure = reported_exposure.groupby(self.groupby).sum()
        trend_array = self._trend(X)
        X_olf_array = self._onlevel(X)
        sw_olf_array = self._onlevel(sample_weight)
        xp = reported_exposure.get_array_module()
        decay_matrix = self.decay ** xp.abs(
            xp.arange(len_orig)[None].T - xp.arange(len_orig)[None]
        )
        weighted_exposure = xp.swapaxes(reported_exposure.values, -1, -2) * decay_matrix
        trended_ultimate = ((latest.values * trend_array * X_olf_array) /
                            (reported_exposure.values * sw_olf_array))
        trended_ultimate = xp.swapaxes(trended_ultimate, -1, -2)
        apriori = (xp.nansum(weighted_exposure * trended_ultimate, -1) /
                   xp.nansum(weighted_exposure, -1))
        apriori_ = reported_exposure.copy()
        apriori_.values = apriori[..., None]
        detrended_apriori_ = apriori_ / trend_array / X_olf_array * sw_olf_array
        return self._set_ult_attr(apriori_), self._set_ult_attr(detrended_apriori_)


    def predict(self, X, sample_weight=None):
        """Predicts the CapeCod ultimate on a new triangle **X**

        Parameters
        ----------
        X: Triangle
            Loss data to which the model will be applied.
        sample_weight: Triangle
            For exposure-based methods, the exposure to be used for predictions

        Returns
        -------
        X_new: Triangle
            Loss data with CapeCod ultimate applied
        """
        if sample_weight is None:
            raise ValueError("sample_weight is required.")
        X_new = X.copy()
        X_new.ldf_ = self.ldf_
        X_new, X_new.ldf_ = self.intersection(X_new, X_new.ldf_)
        apriori_, detrended_apriori_ = self._get_capecod_aprioris(
            X_new, sample_weight
        )
        expectation_ = detrended_apriori_ * sample_weight.values
        X_new = super().predict(X_new,  expectation_)
        X_new.apriori_ = apriori_
        X_new.detrended_apriori_ = detrended_apriori_
        X_new.expectation_ = expectation_
        return X_new

    def _trend(self, X):
        if self.groupby is None:
            if hasattr(X, "trend_"):
                if self.trend != 0:
                    warnings.warn(
                        "CapeCod Trend assumption is ignored when X has a trend_ property."
                    )
                trend_array = X.trend_.latest_diagonal.values
            else:
                trend_array = (X.trend(self.trend) / X).latest_diagonal.values
        else:
            if hasattr(X, "trend_"):
                if self.trend != 0:
                    warnings.warn(
                        "CapeCod Trend assumption is ignored when X has a trend_ property."
                    )
                trend_array = X.trend_.groupby(self.groupby).latest_diagonal.sum().values
            else:
                trend_array = (X.groupby(self.groupby).sum().trend(self.trend) /
                               X.groupby(self.groupby).sum()).latest_diagonal.values
        return trend_array

    def _onlevel(self, X):
        if self.groupby is None:
            if hasattr(X, "olf_"):
                olf_array = X.olf_.values
            else:
                olf_array = 1.0
        else:
            if hasattr(X, "olf_"):
                olf_array = (X.olf_ * X).groupby(self.groupby).sum().values
            else:
                olf_array = 1.0
        return olf_array



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from abc import abstractmethod

import numpy as np
from chainladder.methods.base import MethodBase
from joblib import Parallel, delayed
from sklearn.base import clone
from sklearn.ensemble._base import _fit_single_estimator, _BaseHeterogeneousEnsemble
from sklearn.ensemble._voting import _BaseVoting
from sklearn.utils import Bunch
from sklearn.utils.validation import (_deprecate_positional_args,
                                      check_is_fitted)

from ..core.base import is_chainladder


class _BaseTriangleEnsemble(_BaseHeterogeneousEnsemble):
    """Base class for ensemble of triangle methods."""
    def __init__(self, estimators):
        super().__init__(estimators)

    def _validate_estimators(self):
        if self.estimators is None or len(self.estimators) == 0:
            raise ValueError(
                "Invalid 'estimators' attribute, 'estimators' should be a list"
                " of (string, estimator) tuples."
            )
        names, estimators = zip(*self.estimators)
        # defined by MetaEstimatorMixin in _BaseHeterogeneousEnsemble
        self._validate_names(names)

        for est in estimators:
            if est != 'drop' and not is_chainladder(est):
                raise ValueError(
                    f"The estimator {est.__class__.__name__}"
                    f" should be a chainladder method."
                )

        return names, estimators


class _BaseChainladderVoting(_BaseVoting, _BaseTriangleEnsemble):
    """Base class for voting between chainladder methods.

    Warning: This class should not be used directly. Use derived classes
    instead.
    """
    def _assum_vector_is_none(self, X, assum_vector):
        # return np.ones(X.shape[:3] + (len(self.estimators), ))
        return np.repeat(
            np.array(self.default_weighting)[np.newaxis, :],
            repeats=X.shape[-1],
            axis=0
            )

    def _assum_vector_is_array(self, X, assum_vector):
        assum_vector_ = assum_vector
        return assum_vector_

    def _assum_vector_is_list(self, X, assum_vector):
        assum_vector_ = np.array(assum_vector)
        return assum_vector_

    def _assum_vector_is_callable(self, X, assum_vector):
        assum_vector_ = np.array([*self.X_.origin.map(assum_vector)])
        return assum_vector_

    def _assum_vector_is_dict(self, X, assum_vector):
        mapping_dict = {X.origin[X.origin == k][0]: v for k, v in assum_vector.items()}
        missing = {k: self.default_weighting for k in X.origin[~X.origin.isin(mapping_dict.keys())]}
        assum_vector_ = np.array([*self.X_.origin.map({**mapping_dict, **missing})])
        return assum_vector_

    def _coerce_assum_vector_to_array(self, X, assum_vector):
        if assum_vector is None:
            assum_vector_ = self._assum_vector_is_none(X, assum_vector)
        elif isinstance(assum_vector, np.ndarray):
            assum_vector_ = self._assum_vector_is_array(X, assum_vector)
        elif isinstance(assum_vector, list):
            assum_vector_ = self._assum_vector_is_list(X, assum_vector)
        elif callable(assum_vector):
            assum_vector_ = self._assum_vector_is_callable(X, assum_vector)
        elif isinstance(assum_vector, dict):
            assum_vector_ = self._assum_vector_is_dict(X, assum_vector)
        else:
            raise ValueError("The vector assumption provided must be a "
                             "numpy array, list, dict or callable. Got a "
                             f"{type(assum_vector)} instead.")
        return assum_vector_

    def _broadcast_weights(self, X):
        if self.weights_.ndim == 3:
            weights = np.repeat(self.weights_[np.newaxis, ...], X.shape[1], axis=0)
        if self.weights_.ndim == 4:
            weights = np.repeat(self.weights_[np.newaxis, ...], X.shape[0], axis=0)
        return weights

    def _predict(self, X, sample_weight=None):
        """Collect results from clf.predict calls."""
        return [est.predict(X, sample_weight=sample_weight) for est in self.estimators_]

    @abstractmethod
    def fit(self, X, y, sample_weight=None):
        """Get common fit operations."""
        names, clfs = self._validate_estimators()
        if self.default_weighting is None:
            self.default_weighting = (1, ) * len(self.estimators)
        self.weights_ = self._coerce_assum_vector_to_array(X, self.weights)
        self.weights_ = self.weights_[..., np.newaxis]
        if self.weights_.shape[-3] != X.shape[2]:
            raise ValueError('Length of weight arrays do not equal'
                             f' number of accident periods; found'
                             f' {self.weights_.shape[-3]} weights'
                             f' and {X.shape[2]} accident periods.')
        if self.weights_.shape[-2] != len(self.estimators):
            raise ValueError('Number of weight arrays does not equal'
                             f' number of estimators array; '
                             f' found {self.weights_.shape[-2]} weights'
                             f' arrays and {len(self.estimators)}'
                             ' estimators.')

        self.estimators_ = Parallel(n_jobs=self.n_jobs)(
                delayed(_fit_single_estimator)(
                        clone(clf), X, y,
                        sample_weight=sample_weight,
                        message_clsname='VotingChainladder',
                        message=self._log_message(names[idx],
                                                  idx + 1, len(clfs))
                )
                for idx, clf in enumerate(clfs) if clf != 'drop'
            )

        self.named_estimators_ = Bunch()

        # Uses 'drop' as placeholder for dropped estimators
        est_iter = iter(self.estimators_)
        for name, est in self.estimators:
            current_est = est if est == 'drop' else next(est_iter)
            self.named_estimators_[name] = current_est

        return self


class VotingChainladder(_BaseChainladderVoting, MethodBase):
    """Prediction voting chainladder method for unfitted estimators.

    A voting chainladder is an ensemble meta-estimator that fits several base
    chainladder methods, each on the whole triangle. Then it combines the
    individual predictions based on a matrix of weights to form a
    final prediction.

    Read more in the :ref:`User Guide <voting>`.

    .. versionadded:: 0.8.0

    Parameters
    ----------
    estimators: list of (str, estimator) tuples
        Invoking the ``fit`` method on the ``VotingChainladder`` will fit clones
        of those original estimators that will be stored in the class attribute
        ``self.estimators_``. An estimator can be set to ``'drop'`` using
        ``set_params``.
    weights: array callable or dict, default=None
        ``array``: Numpy array of shape (index, columns, origin, n_estimators). Minimum
        shape required is (origin, n_estimators). Lower dimensional weight arrays
        will have missing dimensions repeated to match the shape of the triangle.
        ``list``: List of weights where each weight is a list of length n_estimators.
        ``dict``: A dictionary where the origin is mapped to a weighting tuple. Missing
        origin periods will be given ``default_weighting``.
        ``callable``: A callable that returns weighting tuples.
        ``None`` uses ``default_weighting``.
    default_weighting: tuple of shape (n_estimators, ), default=None
        Default weighting to use where a weight was not provided or if ``weights`` is None.
        ``None`` uses a typle of all ones which is equivalent to averaging the predictions
        of the estimators.
    n_jobs: int, default=None
        The number of jobs to run in parallel for ``fit``.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors.
        for more details.
    verbose: bool, default=False
        If True, the time elapsed while fitting will be printed as it
        is completed.

    Attributes
    ----------
    estimators_: list of chainladder estimators
        The collection of fitted sub-estimators as defined in ``estimators``
        that are not 'drop'.
    named_estimators_: Bunch
        Attribute to access any fitted sub-estimators by name.

    Examples
    --------
    >>> import numpy as np
    >>> import chainladder as cl
    >>> raa = cl.load_sample('RAA')
    >>> cl_ult = cl.Chainladder().fit(raa).ultimate_  # Chainladder Ultimate
    >>> apriori = cl_ult * 0 + (float(cl_ult.sum()) / 10)  # Mean Chainladder Ultimate
    >>> bcl = cl.Chainladder()
    >>> bf = cl.BornhuetterFerguson()
    >>> cc = cl.CapeCod()
    >>> estimators = [('bcl', bcl), ('bf', bf), ('cc', cc)]
    >>> weights = np.array([[0.6, 0.2, 0.2]] * 4 + [[0, 0.5, 0.5]] * 3 + [[0, 0, 1]] * 3)
    >>> vot = cl.VotingChainladder(estimators=estimators, weights=weights)
    >>> vot.fit(raa, sample_weight=apriori)
    >>> print(vot.ultimate_)
                      2262
        1981  18834.000000
        1982  16875.500226
        1983  24058.534810
        1984  28542.580970
        1985  28236.843134
        1986  19905.317262
        1987  18947.245455
        1988  23106.943030
        1989  20004.502125
        1990  21605.832631
    """
    @_deprecate_positional_args
    def __init__(self, estimators, *, weights=None, default_weighting=None,
                 n_jobs=None, verbose=False):
        super().__init__(estimators=estimators)
        self.weights = weights
        self.default_weighting = default_weighting
        self.n_jobs = n_jobs
        self.verbose = verbose

    def fit(self, X, y=None, sample_weight=None):
        """Fit the estimators.

        Parameters
        ----------
        X : Triangle
            Loss data to which the voting will be applied.

        y : None
            Ignored

        sample_weight : Triangle, default=None
            Exposure to be used in the calculation. Required if any
            of the estimators are exposure based.

        Returns
        -------
        self : object
            Fitted estimator.
        """
        self.X_ = self.validate_X(X)
        super().fit(X, y, sample_weight)
        self.ultimate_ = self._get_ultimate(X, sample_weight=sample_weight)
        return self

    def predict(self, X, sample_weight=None):
        """Predicts the voting chainladder ultimate on a new triangle **X**

        Predicts the ultimate for each of the estimators and combines them
        into a single ultimate based on the weights given.

        Parameters
        ----------
        X : Triangle
            Loss data to which the model will be applied.
        sample_weight : Triangle, default=None
            Exposure to be used in the calculation. Required if any
            of the estimators are exposure based.

        Returns
        -------
        X_new: Triangle
            Loss data with VotingChainladder ultimate applied
        """
        check_is_fitted(self)
        obj = X.copy()
        # obj.ldf_ = self.ldf_
        self.validate_weight(X, sample_weight)
        if sample_weight:
            sample_weight = sample_weight.set_backend(obj.array_backend)
        obj.ultimate_ = self._get_ultimate(obj, sample_weight=sample_weight)
        return obj

    def transform(self, X, sample_weight=None):
        """Return predictions for VotingChainladder

        Parameters
        ----------
        X : Triangle
            Loss data to which the model will be applied.

        sample_weight : Triangle, default=None
            Exposure to be used in the calculation. Required if any
            of the estimators are exposure based.

        Returns
        -------
        X_new: Triangle
            Loss data with VotingChainladder ultimate applied
        """
        return self.predict(X, sample_weight=sample_weight)

    def fit_transform(self, X, y=None, sample_weight=None):
        """Fit and return predictions for VotingChainladder

        Parameters
        ----------
        X : Triangle
            Loss data to which the model will be applied.

        y : None
            Ignored

        sample_weight : Triangle, default=None
            Exposure to be used in the calculation. Required if any
            of the estimators are exposure based.

        Returns
        -------
        X_new: Triangle
            Loss data with VotingChainladder ultimate applied
        """
        return self.fit(X, y, sample_weight).transform(X, sample_weight=sample_weight)

    def _get_ultimate(self, X, sample_weight=None):
        if self.weights_.ndim < 5:
            weights = self._broadcast_weights(X)
        elif self.weights_.ndim == 5:
            if self.weights_.shape[0] != X.shape[0]:
                raise ValueError('Index length (axis 0) of weights does'
                                 ' not equal index length of X; found'
                                 f' {self.weights_.shape[0]} for weights'
                                 f' and {X.shape[0]} for X.')
            if self.weights_.shape[1] != X.shape[1]:
                raise ValueError('Column length (axis 1) of weights does'
                                 f' not equal column length of X; found'
                                 f' {self.weights_.shape[1]} for weights'
                                 f' and {X.shape[1]} for X.')
            weights = self.weights_

        ultimate = sum([
                est.predict(X, sample_weight).ultimate_ * weights[..., i, :]
                for i, est in enumerate(self.estimators_)
            ]) / weights.sum(axis=-2)
        return ultimate



import chainladder as cl
import pytest

def test_grid(clrd):
    # Load Data
    medmal_paid = clrd.groupby("LOB").sum().loc["medmal"]["CumPaidLoss"]
    medmal_prem = (
        clrd.groupby("LOB").sum().loc["medmal"]["EarnedPremDIR"].latest_diagonal
    )

    # Pipeline
    dev = cl.Development()
    tail = cl.TailCurve()
    benk = cl.Benktander()

    steps = [("dev", dev), ("tail", tail), ("benk", benk)]
    pipe = cl.Pipeline(steps)

    # Prep Benktander Grid Search with various assumptions, and a scoring function
    param_grid = dict(benk__n_iters=[250], benk__apriori=[1.00])
    scoring = {"IBNR": lambda x: x.named_steps.benk.ibnr_.sum()}

    grid = cl.GridSearch(pipe, param_grid, scoring=scoring)
    # Perform Grid Search
    grid.fit(medmal_paid, benk__sample_weight=medmal_prem)
    assert (
        grid.results_["IBNR"][0]
        == cl.Benktander(n_iters=250, apriori=1)
        .fit(
            cl.TailCurve().fit_transform(cl.Development().fit_transform(medmal_paid)),
            sample_weight=medmal_prem,
        )
        .ibnr_.sum()
    )


@pytest.fixture
def tri(clrd):
    tri = clrd.groupby('LOB').sum()[['CumPaidLoss', 'IncurLoss', 'EarnedPremDIR']]
    tri['CaseIncurredLoss'] = tri['IncurLoss'] - tri['CumPaidLoss']
    return tri

dev = [cl.Development, cl.ClarkLDF, cl.Trend, cl.IncrementalAdditive,
       lambda : cl.MunichAdjustment(paid_to_incurred=('CumPaidLoss', 'CaseIncurredLoss')),
       lambda :cl.CaseOutstanding(paid_to_incurred=('CumPaidLoss', 'CaseIncurredLoss'))]
tail = [cl.TailCurve, cl.TailConstant, cl.TailBondy, cl.TailClark]
ibnr = [cl.Chainladder,  cl.BornhuetterFerguson,
        lambda : cl.Benktander(n_iters=2), cl.CapeCod]

@pytest.mark.parametrize('dev', dev)
@pytest.mark.parametrize('tail', tail)
@pytest.mark.parametrize('ibnr', ibnr)
def test_pipeline(tri, dev, tail, ibnr):
    X = tri[['CumPaidLoss', 'CaseIncurredLoss']]
    sample_weight = tri['EarnedPremDIR'].latest_diagonal
    cl.Pipeline(
        steps=[('dev', dev()), ('tail', tail()), ('ibnr', ibnr())]
    ).fit_predict(X, sample_weight=sample_weight).ibnr_.sum('origin').sum('columns').sum()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from chainladder.methods.chainladder import Chainladder
from chainladder.development import DevelopmentBase, Development
import numpy as np
import pandas as pd
from warnings import warn
import types


class BootstrapODPSample(DevelopmentBase):
    """
    Class to generate bootstrap samples of triangles.  Currently this Only
    supports 'single' triangles (single index and single column).

    Parameters
    ----------
    n_sims: int (default=1000)
        Number of simulations to generate
    n_periods: integer, optional (default=-1)
        number of origin periods to be used in the ldf average calculation. For
        all origin periods, set n_periods=-1
    hat_adj: bool (default=False)
        Adjust standardized Pearson residuals with the hat matrix adjustment
        factor.  If false, Degree of Freedom adjustment is used.
    drop: tuple or list of tuples
        Drops specific origin/development combination(s) from residual sample
    drop_high: bool or list of bool (default=None)
        Drops highest link ratio(s) from residual sample
    drop_low: bool or list of bool (default=None)
        Drops lowest link ratio(s) from residual sample
    drop_valuation: str or list of str (default=None)
        Drops specific valuation periods from residual sample. str must be date
        convertible.
    random_state: int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by np.random.

    Attributes
    ----------
    resampled_triangles_: Triangle
        A set of triangles represented by each simulation
    scale_:
        The scale parameter to be used in generating process risk
    """

    def __init__(
        self,
        n_sims=1000,
        n_periods=-1,
        hat_adj=True,
        drop=None,
        drop_high=None,
        drop_low=None,
        drop_valuation=None,
        random_state=None,
    ):
        self.n_sims = n_sims
        self.n_periods = n_periods
        self.hat_adj = hat_adj
        self.drop = drop
        self.drop_high = drop_high
        self.drop_low = drop_low
        self.drop_valuation = drop_valuation
        self.random_state = random_state

    def fit(self, X, y=None, sample_weight=None):
        if X.shape[1] > 1:
            from chainladder.utils.utility_functions import concat
            out = [BootstrapODPSample(**self.get_params()).fit(X.iloc[:, i])
                   for i in range(X.shape[1])]
            xp = X.get_array_module(out[0].design_matrix_)
            self.design_matrix_ = xp.concatenate([i.design_matrix_[None] for i in out], axis=0)
            self.hat_ = xp.concatenate([i.hat_[None] for i in out], axis=0)
            self.resampled_triangles_ = concat([i.resampled_triangles_ for i in out], axis=1)
            self.scale_ = xp.array([i.scale_ for i in out])
            self.w_ = out[0].w_
        else:
            backend = X.array_backend
            if backend == "sparse":
                X = X.set_backend("numpy")
            else:
                X = X.copy()
            xp = X.get_array_module()
            if len(X) != 1:
                raise ValueError("Only single index triangles are supported")
            if type(X.ddims) != np.ndarray:
                raise ValueError("Triangle must be expressed with development lags")
            lag = {"M": 1, "Q": 3, "Y": 12}[X.development_grain]
            obj = Development(
                n_periods=self.n_periods,
                drop=self.drop,
                drop_high=self.drop_high,
                drop_low=self.drop_low,
                drop_valuation=self.drop_valuation,
            ).fit_transform(X)
            self.w_ = obj.w_
            obj = Chainladder().fit(obj)
            # Works for only a single triangle - can we generalize this
            exp_incr_triangle = obj.full_expectation_.cum_to_incr().values[
                0, 0, :, : X.shape[-1]
            ]
            exp_incr_triangle = xp.nan_to_num(exp_incr_triangle) * obj.X_.nan_triangle
            self.design_matrix_ = self._get_design_matrix(X)
            if self.hat_adj:
                try:
                    self.hat_ = self._get_hat(X, exp_incr_triangle)
                except:
                    warn("Could not compute hat matrix.  Setting hat_adj to False")
                    self.had_adj = False
                    self.hat_ = None
            else:
                self.hat_ = None
            self.resampled_triangles_, self.scale_ = self._get_simulation(
                X, exp_incr_triangle
            )
            n_obs = xp.nansum(self.w_)
            n_origin_params = X.shape[2]
            n_dev_params = X.shape[3] - 1
            deg_free = n_obs - n_origin_params - n_dev_params
            deg_free_adj_fctr = xp.sqrt(n_obs / deg_free)
        return self

    def _get_simulation(self, X, exp_incr_triangle):
        xp = X.get_array_module()
        k_value = 1  # for ODP Poisson
        unscaled_residuals = (
            (X.cum_to_incr().values - exp_incr_triangle)
            / xp.sqrt(xp.abs(exp_incr_triangle ** k_value))
        )[0, 0, ...]
        w_ = self.w_[0, 0]
        w_[:, 1:] * w_[:, 1:]
        w_ = xp.concatenate((w_[:, 0:1], w_), axis=1)
        unscaled_residuals = unscaled_residuals * w_
        pearson_chi_sq = sum(sum(xp.nan_to_num(unscaled_residuals) ** 2))
        if self.hat_ is None:
            standardized_residuals = unscaled_residuals
        else:
            standardized_residuals = self.hat_ * unscaled_residuals
        n_params = self.design_matrix_.shape[1]
        degree_freedom = xp.nansum(X.nan_triangle) - n_params
        # Shapland has a hetero adjustment to degree_freedom here
        # He also adjusts the residuals for the hetero adjustment
        scale_phi = pearson_chi_sq / degree_freedom
        k, v, o, d = X.shape
        resids = xp.reshape(standardized_residuals, (k, v, o * d))

        adj_resid_dist = resids[xp.isfinite(resids)]  # Missing k,v dimensions
        # Suggestions from Using the ODP Bootstrap Model: A Practitioners Guide
        adj_resid_dist = adj_resid_dist[adj_resid_dist != 0]
        adj_resid_dist = adj_resid_dist - xp.mean(adj_resid_dist)
        random_state = xp.random.RandomState(self.random_state)
        resampled_residual = [
            (
                random_state.choice(
                    adj_resid_dist, size=exp_incr_triangle.shape, replace=True
                )
                * (exp_incr_triangle * 0 + 1)
            )[None, ...]
            for item in range(self.n_sims)
        ]
        resampled_residual = xp.concatenate(tuple(resampled_residual), 0).reshape(
            self.n_sims, exp_incr_triangle.shape[0], exp_incr_triangle.shape[1]
        )
        resampled_residual = resampled_residual
        b = xp.repeat(exp_incr_triangle[None, ...], self.n_sims, 0)
        resampled_triangles = (resampled_residual * xp.sqrt(abs(b)) + b).cumsum(2)
        resampled_triangles = xp.swapaxes(resampled_triangles[None, ...], 0, 1)
        obj = X.copy()
        obj.kdims = np.arange(self.n_sims)
        obj.values = resampled_triangles
        obj._set_slicers()
        return obj, scale_phi

    def _get_design_matrix(self, X):
        """ The design matrix used in hat matrix adjustment (Shapland eq3.12)
        """
        xp = X.get_array_module()
        w = X.nan_triangle
        arr = xp.diag(w[:, 0])
        intra_beta = xp.zeros((w.shape[0], w.shape[1] - 1))
        arr = xp.concatenate((arr, intra_beta), axis=1)
        for i in range(w.shape[1] - 1):
            len_alpha = len(w[:, i + 1][~xp.isnan(w[:, i + 1])])
            intra_alpha = xp.diag(w[:, i + 1])[:len_alpha, :]
            intra_beta[:, i] = 1
            intra_beta = intra_beta[:len_alpha, :]
            intra_arr = xp.concatenate((intra_alpha, intra_beta), axis=1)
            arr = xp.concatenate((arr, intra_arr), axis=0)
        return arr

    def _get_hat(self, X, exp_incr_triangle):
        """ The hat matrix adjustment (Shapland eq3.23)"""
        xp = X.get_array_module()
        weight_matrix = xp.diag(
            pd.DataFrame(exp_incr_triangle).unstack().dropna().values
        )
        design_matrix = self.design_matrix_
        hat = xp.matmul(
            xp.matmul(
                xp.matmul(
                    design_matrix,
                    xp.linalg.inv(
                        xp.matmul(
                            design_matrix.T, xp.matmul(weight_matrix, design_matrix)
                        )
                    ),
                ),
                design_matrix.T,
            ),
            weight_matrix,
        )
        hat = xp.diagonal(xp.sqrt(xp.divide(1, abs(1 - hat), where=(1 - hat) != 0)))
        total_length = X.nan_triangle.shape[0]
        reshaped_hat = xp.reshape(hat[:total_length], (1, total_length))
        indices = xp.nansum(X.nan_triangle, axis=0).cumsum().astype(int)
        for num, item in enumerate(indices[:-1]):
            col_length = int(indices[num + 1] - indices[num])
            col = xp.reshape(
                hat[int(indices[num]) : int(indices[num + 1])], (1, col_length)
            )
            nans = xp.repeat(
                xp.array([xp.nan])[None, :], total_length - col_length, axis=1
            )
            col = xp.concatenate((col, nans), axis=1)
            reshaped_hat = xp.concatenate((reshaped_hat, col), axis=0)
        return reshaped_hat.T

    def _get_hetero_adjustment(self):
        pass

    def transform(self, X):
        """ If X and self are of different shapes, align self to X, else
        return self.

        Parameters
        ----------
        X: Triangle
            The triangle to be transformed

        Returns
        -------
            X_new: New triangle with transformed attributes.
        """
        X_new = X.copy()
        X_new = self.resampled_triangles_
        X_new.scale_ = self.scale_
        X_new.random_state = self.random_state
        X_new._get_process_variance = types.MethodType(_get_process_variance, X_new)
        return X_new


def _get_process_variance(self, full_triangle):
    """ Inserts random noise into the full triangles and full expectations """
    # if self.process_dist == 'od poisson':
    #    process_triangle = np.nan_to_num(np.array([random_state.poisson(lam=abs(item))*np.sign(np.nan_to_num(item))for item in sim_exp_incr_triangle]))
    xp = full_triangle.get_array_module()
    lower_tri = full_triangle.cum_to_incr() - self.cum_to_incr()
    random_state = xp.random.RandomState(
        None if not self.random_state else self.random_state + 1
    )
    lower_tri.values = random_state.gamma(
        shape=abs(lower_tri.values) / self.scale_, scale=self.scale_
    ) * xp.sign(xp.nan_to_num(lower_tri.values))
    return (lower_tri + self.cum_to_incr()).incr_to_cum()



import chainladder as cl
import numpy as np
from chainladder.utils.cupy import cp


def test_constant_balances(qtr):
    xp = qtr.get_array_module()
    assert (
        round(
            float(
                xp.prod(
                    cl.TailConstant(1.05, decay=0.8)
                    .fit(qtr)
                    .ldf_.iloc[0, 1]
                    .values[0, 0, 0, -5:]
                )
            ),
            3,
        )
        == 1.050
    )
